/*
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: buffer.c,v 1.3 2000/11/10 02:38:50 guenter Exp $
 *
 *
 * contents:
 *
 * buffer_entry structure - serves as a transport encapsulation
 *   of the mpeg audio/video data through xine
 *
 * free buffer pool management routines
 *
 * FIFO buffer structures/routines
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include "buffer.h"

/*
 * global variables
 */

buf_element_t   *gBufferPoolTop;    /* a stack actually */
pthread_mutex_t  gBufferPoolMutex;
pthread_cond_t   gBufferPoolCondNotEmpty;
int              gBufferPoolNumFree;

void buffer_pool_free (buf_element_t *pBufElement) {

  pthread_mutex_lock (&gBufferPoolMutex);

  pBufElement->next = gBufferPoolTop;
  gBufferPoolTop = pBufElement;

  gBufferPoolNumFree++;

  pthread_cond_signal (&gBufferPoolCondNotEmpty);

  pthread_mutex_unlock (&gBufferPoolMutex);
}


void buffer_pool_init (int nNumBuffers, uint32_t buf_size) {

  int i;
  
  gBufferPoolTop = NULL;

  pthread_mutex_init (&gBufferPoolMutex, NULL);
  pthread_cond_init (&gBufferPoolCondNotEmpty, NULL);

  for (i = 0; i<nNumBuffers; i++) {
    buf_element_t *pBuf;

    pBuf = malloc (sizeof (buf_element_t));

    pBuf->pMem     = malloc (buf_size);
    pBuf->nMaxSize = buf_size;
    
    buffer_pool_free (pBuf);
  }
  gBufferPoolNumFree = nNumBuffers;
}

int buffer_pool_isempty () {
  return gBufferPoolNumFree<7;
}

buf_element_t *buffer_pool_alloc () {
  
  buf_element_t *pBuf;

  pthread_mutex_lock (&gBufferPoolMutex);

  while (!gBufferPoolTop) {
    pthread_cond_wait (&gBufferPoolCondNotEmpty, &gBufferPoolMutex);
  }

  pBuf = gBufferPoolTop;
  gBufferPoolTop = gBufferPoolTop->next;
  gBufferPoolNumFree--;

  pthread_mutex_unlock (&gBufferPoolMutex);

  return pBuf;
}

fifo_buffer_t *fifo_buffer_new () {

  fifo_buffer_t *pFifo;

  pFifo = malloc (sizeof (fifo_buffer_t));

  pFifo->mpFirst = NULL;
  pFifo->mpLast  = NULL;

  pthread_mutex_init (&pFifo->mMutex, NULL);
  pthread_cond_init (&pFifo->mNotEmpty, NULL);

  return pFifo;
}

void fifo_buffer_put (fifo_buffer_t *pFifo, buf_element_t *pBufElement) {
  
  pthread_mutex_lock (&pFifo->mMutex);

  if (pFifo->mpLast) 
    pFifo->mpLast->next = pBufElement;
  else 
    pFifo->mpFirst = pBufElement;

  pFifo->mpLast  = pBufElement;
  pBufElement->next = NULL;

  pthread_cond_signal (&pFifo->mNotEmpty);

  pthread_mutex_unlock (&pFifo->mMutex);
}

buf_element_t *fifo_buffer_get (fifo_buffer_t *pFifo) {

  buf_element_t *pBuf;
  
  pthread_mutex_lock (&pFifo->mMutex);

  while (pFifo->mpFirst==NULL) {
    pthread_cond_wait (&pFifo->mNotEmpty, &pFifo->mMutex);
  }

  pBuf = pFifo->mpFirst;

  pFifo->mpFirst = pFifo->mpFirst->next;
  if (pFifo->mpFirst==NULL)
    pFifo->mpLast = NULL;

  pthread_mutex_unlock (&pFifo->mMutex);

  return pBuf;
}

int fifo_buffer_is_empty (fifo_buffer_t *pFifo) {
  return (pFifo->mpFirst == NULL);
}

void fifo_buffer_clear (fifo_buffer_t *pFifo) {
  
  buf_element_t *pBuf;

  pthread_mutex_lock (&pFifo->mMutex);

  while (pFifo->mpFirst != NULL) {

    pBuf = pFifo->mpFirst;

    pFifo->mpFirst = pFifo->mpFirst->next;
    if (pFifo->mpFirst==NULL)
      pFifo->mpLast = NULL;

    buffer_pool_free (pBuf);
  }

  pthread_mutex_unlock (&pFifo->mMutex);
}
